@@ -1,4 +1,4 @@
-#  Copyright (c) 2018-2024 by Rocky Bernstein
+#  Copyright (c) 2018-2025 by Rocky Bernstein
 #
 #  This program is free software; you can redistribute it and/or
 #  modify it under the terms of the GNU General Public License
@@ -27,7 +27,7 @@ import sys
 from io import StringIO
 from linecache import getline
 from types import CodeType
-from typing import Iterable, Optional, Union
+from typing import Iterable, Iterator, Optional, Tuple, Union
 
 from xdis.cross_dis import (
     format_code_info,
@@ -174,7 +174,7 @@ def offset2line(offset: int, linestarts):
     return linestarts[high][1]
 
 
-def _parse_varint(iterator):
+def _parse_varint(iterator: Iterator[int]) -> int:
     b = next(iterator)
     val = b & 63
     while b & 64:
@@ -223,8 +223,9 @@ def prefer_double_quote(string: str) -> str:
     return string
 
 
-def get_instructions_bytes(
+def get_logical_instruction_at_offset(
     bytecode,
+    offset: int,
     opc,
     varnames=None,
     names=None,
@@ -233,21 +234,23 @@ def get_instructions_bytes(
     linestarts=None,
     line_offset=0,
     exception_entries=None,
+    labels=None,
 ):
-    """Iterate over the instructions in a bytecode string.
-
-    Generates a sequence of Instruction namedtuples giving the details of each
-    opcode.  Additional information about the code's runtime environment
-    e.g., variable names, constants, can be specified using optional
-    arguments.
-
     """
-    labels = opc.findlabels(bytecode, opc)
-
-    if exception_entries is not None:
-        for start, end, target, _, _ in exception_entries:
-            for i in range(start, end):
-                labels.append(target)
+    Return a single logical instruction for `bytecode` at offset `offset`.
+    if the opcode at offset is EXTENDED_ARG, then instructions are returned
+    until we no longer have an EXTENDED_ARG instruction. Note that the
+    last non-EXTENDED_ARG instruction will have its argument value adjusted
+    to note the increased size of the argument.
+    """
+    if labels is None:
+        labels = opc.findlabels(bytecode, opc)
+        # PERFORMANCE FIX: Only add exception labels if we're building labels ourselves
+        # When called from get_instructions_bytes, labels already includes exception targets
+        if exception_entries is not None:
+            for start, end, target, _, _ in exception_entries:
+                if target not in labels:
+                    labels.append(target)
 
     # label_maps = get_jump_target_maps(bytecode, opc)
 
@@ -256,40 +259,45 @@ def get_instructions_bytes(
     python_36 = True if opc.python_version >= (3, 6) else False
 
     starts_line = None
-    # enumerate() is not an option, since we sometimes process
-    # multiple elements on a single pass through the loop
+
     n = len(bytecode)
-    i = 0
+
     extended_arg_count = 0
     extended_arg = 0
+
     if hasattr(opc, "EXTENDED_ARG"):
         extended_arg_size = instruction_size(opc.EXTENDED_ARG, opc)
     else:
         extended_arg_size = 0
 
-    while i < n:
+    # This is not necessarily true initially, but it gets us through the
+    # loop below.
+
+    last_op_was_extended_arg = True
+    i = offset
+
+    while i < n and last_op_was_extended_arg:
         op = code2num(bytecode, i)
+        opname = opc.opname[op]
+        optype = get_optype(op, opc)
 
         offset = i
         if linestarts is not None:
             starts_line = linestarts.get(i, None)
             if starts_line is not None:
                 starts_line += line_offset
-        if i in labels:
-            is_jump_target = True
-        else:
-            is_jump_target = False
+
+        is_jump_target = i in labels
 
         i += 1
         arg = None
         argval = None
         argrepr = ""
         has_arg = op_has_argument(op, opc)
-        optype = get_optype(op, opc)
         if has_arg:
             if python_36:
                 arg = code2num(bytecode, i) | extended_arg
-                extended_arg = (arg << 8) if op == opc.EXTENDED_ARG else 0
+                extended_arg = (arg << 8) if opname == "EXTENDED_ARG" else 0
                 # FIXME: Python 3.6.0a1 is 2, for 3.6.a3 we have 1
                 i += 1
             else:
@@ -299,11 +307,7 @@ def get_instructions_bytes(
                     + extended_arg
                 )
                 i += 2
-                extended_arg = (
-                    arg * 0x10000
-                    if hasattr(opc, "EXTENDED_ARG") and op == opc.EXTENDED_ARG
-                    else 0
-                )
+                extended_arg = arg * 0x10000 if opname == "EXTENDED_ARG" else 0
 
             #  Set argval to the dereferenced value of the argument when
             #  available, and argrepr to the string representation of argval.
@@ -311,67 +315,83 @@ def get_instructions_bytes(
             #    raw name index for LOAD_GLOBAL, LOAD_CONST, etc.
 
             argval = arg
+
+            # create a localsplusnames table that resolves duplicates.
+            localsplusnames = (varnames or tuple()) + tuple(
+                name for name in (cells or tuple()) if name not in varnames
+            )
+
             if op in opc.CONST_OPS:
                 argval, argrepr = _get_const_info(arg, constants)
             elif op in opc.NAME_OPS:
-                if opc.version_tuple >= (3, 11) and opc.opname[op] == "LOAD_GLOBAL":
+                if opc.version_tuple >= (3, 11) and opname == "LOAD_GLOBAL":
                     argval, argrepr = _get_name_info(arg >> 1, names)
                     if arg & 1:
                         argrepr = "NULL + " + argrepr
-                elif opc.version_tuple >= (3, 12) and opc.opname[op] == "LOAD_ATTR":
+                elif opc.version_tuple >= (3, 12) and opname == "LOAD_ATTR":
                     argval, argrepr = _get_name_info(arg >> 1, names)
                     if arg & 1:
                         argrepr = "NULL|self + " + argrepr
-                elif (
-                    opc.version_tuple >= (3, 12) and opc.opname[op] == "LOAD_SUPER_ATTR"
-                ):
+                elif opc.version_tuple >= (3, 12) and opname == "LOAD_SUPER_ATTR":
                     argval, argrepr = _get_name_info(arg >> 2, names)
                     if arg & 1:
                         argrepr = "NULL|self + " + argrepr
                 else:
                     argval, argrepr = _get_name_info(arg, names)
             elif op in opc.JREL_OPS:
-                signed_arg = -arg if "JUMP_BACKWARD" in opc.opname[op] else arg
+                signed_arg = -arg if "JUMP_BACKWARD" in opname else arg
                 argval = i + get_jump_val(signed_arg, opc.python_version)
+
+                # check cache instructions for python 3.13
+                if opc.version_tuple >= (3, 13):
+                    if opc.opname[op] in [
+                        "POP_JUMP_IF_TRUE",
+                        "POP_JUMP_IF_FALSE",
+                        "POP_JUMP_IF_NONE",
+                        "POP_JUMP_IF_NOT_NONE",
+                        "JUMP_BACKWARD",
+                    ]:
+                        argval += 2
+
                 # FOR_ITER has a cache instruction in 3.12
-                if opc.version_tuple >= (3, 12) and opc.opname[op] == "FOR_ITER":
+                if opc.version_tuple >= (3, 12) and opname == "FOR_ITER":
                     argval += 2
                 argrepr = "to " + repr(argval)
             elif op in opc.JABS_OPS:
                 argval = get_jump_val(arg, opc.python_version)
                 argrepr = "to " + repr(argval)
             elif op in opc.LOCAL_OPS:
-                if opc.version_tuple >= (3, 13) and opc.opname[op] in ("LOAD_FAST_LOAD_FAST", "STORE_FAST_LOAD_FAST", "STORE_FAST_STORE_FAST"):
+                if opc.version_tuple >= (3, 13) and opname in (
+                    "LOAD_FAST_LOAD_FAST",
+                    "STORE_FAST_LOAD_FAST",
+                    "STORE_FAST_STORE_FAST",
+                ):
                     arg1 = arg >> 4
                     arg2 = arg & 15
-                    argval1, argrepr1 = _get_name_info(arg1, (varnames or tuple()) + (cells or tuple()))
-                    argval2, argrepr2 = _get_name_info(arg2, (varnames or tuple()) + (cells or tuple()))
+                    argval1, argrepr1 = _get_name_info(arg1, localsplusnames)
+                    argval2, argrepr2 = _get_name_info(arg2, localsplusnames)
                     argval = argval1, argval2
                     argrepr = argrepr1 + ", " + argrepr2
                 elif opc.version_tuple >= (3, 11):
-                    argval, argrepr = _get_name_info(
-                        arg, (varnames or tuple()) + (cells or tuple())
-                    )
+                    argval, argrepr = _get_name_info(arg, localsplusnames)
                 else:
                     argval, argrepr = _get_name_info(arg, varnames)
             elif op in opc.FREE_OPS:
                 if opc.version_tuple >= (3, 11):
-                    argval, argrepr = _get_name_info(
-                        arg, (varnames or tuple()) + (cells or tuple())
-                    )
+                    argval, argrepr = _get_name_info(arg, localsplusnames)
                 else:
                     argval, argrepr = _get_name_info(arg, cells)
             elif op in opc.COMPARE_OPS:
-                if opc.python_version >= (3,13):
+                if opc.python_version >= (3, 13):
                     # The fifth-lowest bit of the oparg now indicates a forced conversion to bool.
-                    argval = (opc.cmp_op[arg >> 5])
-                elif opc.python_version >= (3,12):
-                    argval = (opc.cmp_op[arg >> 4])
+                    argval = opc.cmp_op[arg >> 5]
+                elif opc.python_version >= (3, 12):
+                    argval = opc.cmp_op[arg >> 4]
                 else:
-                    argval = (opc.cmp_op[arg])
+                    argval = opc.cmp_op[arg]
                 argrepr = argval
             elif op in opc.NARGS_OPS:
-                opname = opc.opname[op]
+                opname = opname
                 if python_36 and opname in ("CALL_FUNCTION", "CALL_FUNCTION_EX"):
                     if opname == "CALL_FUNCTION":
                         argrepr = format_CALL_FUNCTION(code2num(bytecode, i - 1))
@@ -387,17 +407,15 @@ def get_instructions_bytes(
                             code2num(bytecode, i - 2),
                             code2num(bytecode, i - 1),
                         )
-            if hasattr(opc, "opcode_arg_fmt") and opc.opname[op] in opc.opcode_arg_fmt:
-                argrepr = opc.opcode_arg_fmt[opc.opname[op]](arg)
+            if hasattr(opc, "opcode_arg_fmt") and opname in opc.opcode_arg_fmt:
+                argrepr = opc.opcode_arg_fmt[opname](arg)
         else:
             if python_36:
                 i += 1
-            if hasattr(opc, "opcode_arg_fmt") and opc.opname[op] in opc.opcode_arg_fmt:
-                argrepr = opc.opcode_arg_fmt[opc.opname[op]](arg)
+            if hasattr(opc, "opcode_arg_fmt") and opname in opc.opcode_arg_fmt:
+                argrepr = opc.opcode_arg_fmt[opname](arg)
 
-        opname = opc.opname[op]
         inst_size = instruction_size(op, opc) + (extended_arg_count * extended_arg_size)
-        # fallthrough = op not in opc.nofollow
         start_offset = offset if opc.oppop[op] == 0 else None
 
         yield Instruction(
@@ -419,11 +437,9 @@ def get_instructions_bytes(
             start_offset=start_offset,
         )
         # fallthrough
-        extended_arg_count = (
-            extended_arg_count + 1
-            if hasattr(opc, "EXTENDED_ARG") and op == opc.EXTENDED_ARG
-            else 0
-        )
+        last_op_was_extended_arg = True if opname == "EXTENDED_ARG" else False
+        extended_arg_count = extended_arg_count + 1 if last_op_was_extended_arg else 0
+        # end loop
 
 
 def next_offset(op: int, opc, offset: int) -> int:
@@ -434,6 +450,62 @@ def next_offset(op: int, opc, offset: int) -> int:
     return offset + instruction_size(op, opc)
 
 
+def get_instructions_bytes(
+    bytecode,
+    opc,
+    varnames=None,
+    names=None,
+    constants=None,
+    cells=None,
+    linestarts=None,
+    line_offset=0,
+    exception_entries=None,
+):
+    """
+    Iterate over the instructions in a bytecode string.
+
+    Generates a sequence of Instruction namedtuples giving the details of each
+    opcode.  Additional information about the code's runtime environment
+    e.g., variable names, constants, can be specified using optional
+    arguments.
+    """
+    labels = opc.findlabels(bytecode, opc)
+
+    # PERFORMANCE FIX: Build exception labels ONCE, not on every iteration
+    # The old code was O(n^2) because it rebuilt the same list every call to
+    # get_logical_instruction_at_offset
+    if exception_entries is not None:
+        for start, end, target, _, _ in exception_entries:
+            # Only add the target offset, not every offset in the range
+            # This matches what get_logical_instruction_at_offset expects
+            if target not in labels:
+                labels.append(target)
+
+    n = len(bytecode)
+    offset = 0
+
+    while offset < n:
+        instructions = list(
+            get_logical_instruction_at_offset(
+                bytecode,
+                offset,
+                opc,
+                varnames=varnames,
+                names=names,
+                constants=constants,
+                cells=cells,
+                linestarts=linestarts,
+                line_offset=0,
+                exception_entries=exception_entries,
+                labels=labels
+            )
+        )
+
+        for instruction in instructions:
+            yield instruction
+        offset = next_offset(instruction.opcode, opc, instruction.offset)
+
+
 class Bytecode:
     """Bytecode operations involving a Python code object.
 
@@ -443,7 +515,7 @@ class Bytecode:
     Iterating over these yields the bytecode operations as Instruction instances.
     """
 
-    def __init__(self, x, opc, first_line=None, current_offset=None, dup_lines=True):
+    def __init__(self, x, opc, first_line=None, current_offset=None, dup_lines: bool=True) -> None:
         self.codeobj = co = get_code_object(x)
         self._line_offset = 0
         self._cell_names = ()
@@ -464,7 +536,7 @@ class Bytecode:
         self.opnames = opc.opname
         self.current_offset = current_offset
 
-        if opc.version_tuple >= (3, 11) and hasattr(co, "co_exceptiontable"):
+        if opc.version_tuple >= (3, 11) and not opc.is_pypy and hasattr(co, "co_exceptiontable"):
             self.exception_entries = parse_exception_table(co.co_exceptiontable)
         else:
             self.exception_entries = None
@@ -483,7 +555,7 @@ class Bytecode:
             exception_entries=self.exception_entries,
         )
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"{self.__class__.__name__}({self._original_object!r})"
 
     @classmethod
@@ -497,11 +569,11 @@ class Bytecode:
             tb.tb_frame.f_code, opc=opc, first_line=None, current_offset=tb.tb_lasti
         )
 
-    def info(self):
+    def info(self) -> str:
         """Return formatted information about the code object."""
         return format_code_info(self.codeobj, self.opc.version_tuple)
 
-    def dis(self, asm_format="classic", show_source=False):
+    def dis(self, asm_format: str="classic", show_source: bool=False) -> str:
         """Return a formatted view of the bytecode operations."""
         co = self.codeobj
         filename = co.co_filename
@@ -543,7 +615,7 @@ class Bytecode:
         )
         return output.getvalue()
 
-    def distb(self, tb=None):
+    def distb(self, tb=None) -> None:
         """Disassemble a traceback (default: last traceback)."""
         if tb is None:
             try:
@@ -576,7 +648,7 @@ class Bytecode:
         show_lineno = line_starts is not None or self.opc.version_tuple < (2, 3)
         show_source = show_source and show_lineno and first_line_number and filename
 
-        def show_source_text(line_number: Optional[int]):
+        def show_source_text(line_number: Optional[int]) -> None:
             """
             Show the Python source text if all conditions are right:
               * source text was requested - this implies other checks
@@ -767,7 +839,7 @@ class Bytecode:
         )
 
 
-def list2bytecode(inst_list: Iterable, opc, varnames, consts):
+def list2bytecode(inst_list: Iterable, opc, varnames: str, consts: Tuple[None, int]) -> bytes:
     """Convert list/tuple of list/tuples to bytecode
     _names_ contains a list of name objects
     """
@@ -813,7 +885,7 @@ if __name__ == "__main__":
         ("RETURN_VALUE",),
     ]
 
-    def f():
+    def f() -> int:
         a = 2
         return a
 
